{-# OPTIONS_GHC -Wall #-}
module Proof where

import qualified Data.Set as S 
import Control.Monad
import Control.Applicative

import Formula

data Proof = Id Sequent
           | Fix Sequent Proof Proof
           | PrePost Sequent Proof Proof
           | Zero Sequent Proof
           | ImpR Sequent Proof
           | ImpL Sequent Proof Proof
           | NotL Sequent Proof
           | NotR Sequent Proof
           | WeakR Sequent Proof
           | AndR Sequent Proof Proof
           | AndL Sequent Proof
           | OrL Sequent Proof Proof
           | OrR1 Sequent Proof
           | OrR2 Sequent Proof
           | BotL Sequent
           | TopR Sequent
    deriving (Show, Eq)

-- | A result monad.
data Result a
    = NoResult
    | Timeout
    | Result a
    deriving (Show, Eq)

-- | Propagate 'NoResult' and 'Timeout'.
instance Functor Result where
    fmap _ NoResult   = NoResult
    fmap _ Timeout    = Timeout
    fmap f (Result x) = Result (f x)

-- | Priority: Result > NoResult > Timeout
instance Applicative Result where
    pure = Result
    Result f <*> Result x = Result (f x)
    NoResult <*> _        = NoResult
    _        <*> NoResult = NoResult
    _        <*> _        = Timeout

instance Monad Result where
    return = Result
    NoResult >>= _ = NoResult
    Timeout  >>= _ = Timeout
    Result x >>= f = f x

-- | Priority: Result > Timeout > NoResult
instance MonadPlus Result where
    mzero = NoResult
    x@(Result _) `mplus` _            = x
    _            `mplus` x@(Result _) = x
    _            `mplus` Timeout      = Timeout
    Timeout      `mplus` _            = Timeout
    _            `mplus` _            = NoResult

-- * Tactics correspond to sequent rules

type Cost = Int
type Tactic = Sequent -> S.Set Sequent -> Cost -> Result Proof

tacId :: Tactic
tacId s@(S pre f) _ _ = if f `S.member` pre then return (Id s) else mzero

tacZero :: Tactic
tacZero s@(S pre (Coimp _ q)) seen c = Zero s <$> proof (S pre q) seen c
tacZero _ _ _ = mzero

tacPrePost :: Tactic
tacPrePost s@(S pre (Coimp p' q')) seen c =
    msum [ PrePost s <$> proof (S (S.insert p' pre) p ) seen c
                     <*> proof (S (S.insert q  pre) q') seen c
           | Coimp p q <- S.toList pre ]
tacPrePost _ _ _ = mzero

tacFix :: Tactic
tacFix s@(S pre r) seen c =
    msum [ Fix s <$> proof (S (S.insert r pre) p) seen c
                 <*> proof (S (S.insert q pre) r) seen c
           | Coimp p q <- S.toList pre ]

{-
tacFix' :: Tactic
tacFix' s@(S pre r) seen c =
    msum [ Fix s <$> proof (S (S.insert q pre) p) seen c
                 <*> proof (S (S.insert q pre) r) seen c
           | Coimp p q <- S.toList pre ]
-}

tacImpR :: Tactic
tacImpR s@(S pre (Imp p q)) seen c =
    ImpR s <$> proof (S (S.insert p pre) q) seen c
tacImpR _ _ _ = mzero

tacImpL :: Tactic
tacImpL s@(S pre r) seen c =
    msum [ ImpL s <$> proof (S pre              p) seen c
                  <*> proof (S (S.insert q pre) r) seen c
           | Imp p q <- S.toList pre ]

tacNotL :: Tactic
tacNotL s@(S pre _) seen c =  -- FIXME F vs _any
    msum [ NotL s <$> proof (S pre p) seen c
           | Not p <- S.toList pre ]
-- tacNotL _ _ _ = mzero


tacNotR :: Tactic
tacNotR s@(S pre (Not p)) seen c =
    let ps = S.insert p pre
        false = F -- FIXME
                -- genFresh (S.insert p pre) "bot"
        in NotR s <$> proof (S ps false) seen c
tacNotR _ _ _ = mzero

{- Not needed.
tacWeakR :: Tactic
tacWeakR s@(S pre _) seen c =
    WeakR s <$> proof (S pre F) seen c
-}

tacAndL :: Tactic
tacAndL s@(S pre r) seen c =
    msum [ AndL s <$> proof (S (S.insert q (S.insert p pre)) r) seen c
           | And p q <- S.toList pre ]

tacAndR :: Tactic
tacAndR s@(S pre (And p q)) seen c =
           AndR s <$> proof (S pre p) seen c
                  <*> proof (S pre q) seen c
tacAndR _ _ _ = mzero

tacOrL :: Tactic
tacOrL s@(S pre r) seen c =
    msum [ OrL s <$> proof (S (S.insert p pre) r) seen c
                 <*> proof (S (S.insert q pre) r) seen c
           | Or p q <- S.toList pre ]

tacOrR :: Tactic
tacOrR s@(S pre (Or p q)) seen c =
    msum [ OrR1 s <$> proof (S pre p) seen c
         , OrR2 s <$> proof (S pre q) seen c ]
tacOrR _ _ _ = mzero

tacBotL :: Tactic
tacBotL s@(S pre _) _ _ = if F `S.member` pre then return $ BotL s else mzero

tacTopR :: Tactic
tacTopR s@(S _ T) _ _ = return $ TopR s
tacTopR _ _ _ = mzero

-- TODO costs should be tuned...
tactics :: [ (Tactic, Cost) ]
tactics = [ (tacId, 1)
          , (tacImpL, 1)
          , (tacZero, 1), (tacPrePost, 1), (tacFix, 1)
          , (tacImpR, 1)
          , (tacNotL, 1), (tacNotR, 1)
          , (tacAndL, 1), (tacAndR, 1)
          , (tacOrL, 1), (tacOrR, 1)
          , (tacBotL, 1), (tacTopR, 1)
          ]

-- | Search for a 'Proof', within the given timeout.
proof :: Tactic
proof s seen c 
      | c <= 0            = Timeout
      | s `S.member` seen = mzero    -- loop detected
      | otherwise         = msum [ t s (S.insert s seen) (c-c')
                                   | (t,c') <- tactics ]

-- | Try 'proof' many times, increasing the threshold.
proofSearch :: Formula -> [Result Proof]
proofSearch p = map (proof (S S.empty p) S.empty) [1..]

-- | Decide the 'Formula'.
decide :: Formula -> Result Proof
decide = head . dropWhile (== Timeout) . proofSearch

-- | As 'decide', but show progress.
decideWithProgress :: (Int -> IO ()) -> Formula -> IO (Result Proof)
decideWithProgress act = go 1 . proofSearch
    where go n (Timeout : xs) = act n >> go (succ n) xs
          go n (answer  : _ ) = act n >> return answer
          go _ []             = error "decideWithProgress"

-- * Pretty printer

-- FIXME Highly inefficent
prettyP :: Proof -> String
prettyP (Id s) = "[Id: " ++ prettyS s ++ "]"
prettyP (Fix s p1 p2) =
    "[Fix: " ++ prettyS s ++ " " ++ prettyP p1 ++ " " ++ prettyP p2 ++ "]"
prettyP (PrePost s p1 p2) = 
    "[Prepost: " ++ prettyS s ++ " " ++ prettyP p1 ++ " " ++ prettyP p2 ++ "]"
prettyP (Zero s p) = 
    "[Zero: " ++ prettyS s ++ " " ++ prettyP p ++ "]"
prettyP (ImpR s p) = 
    "[ImpR: " ++ prettyS s ++ " " ++ prettyP p ++ "]"
prettyP (ImpL s p1 p2) = 
    "[ImpL: " ++ prettyS s ++ " " ++ prettyP p1 ++ " " ++ prettyP p2 ++ "]"
prettyP (NotL s p) = 
    "[NotL: " ++ prettyS s ++ " " ++ prettyP p ++ "]"
prettyP (NotR s p) = 
    "[NotR: " ++ prettyS s ++ " " ++ prettyP p ++ "]"
prettyP (WeakR s p) = 
    "[WeakR: " ++ prettyS s ++ " " ++ prettyP p ++ "]"
prettyP (AndR s p1 p2) = 
    "[AndR: " ++ prettyS s ++ " " ++ prettyP p1 ++ " " ++ prettyP p2 ++ "]"
prettyP (AndL s p) = 
    "[AndL: " ++ prettyS s ++ " " ++ prettyP p ++ "]"
prettyP (OrL s p1 p2) = 
    "[OrL: " ++ prettyS s ++ " " ++ prettyP p1 ++ " " ++ prettyP p2 ++ "]"
prettyP (OrR1 s p) = 
    "[OrR1: " ++ prettyS s ++ " " ++ prettyP p ++ "]"
prettyP (OrR2 s p) = 
    "[OrR2: " ++ prettyS s ++ " " ++ prettyP p ++ "]"
prettyP (BotL s) =
    "[BotL: " ++ prettyS s ++ "]"
prettyP (TopR s) =
    "[TopR: " ++ prettyS s ++ "]"

prettyR :: Result Proof -> String
prettyR NoResult   = "NoResult"
prettyR Timeout    = "Timeout"
prettyR (Result x) = "Result " ++ prettyP x

-- * LaTeX pretty printer

prettyL :: Proof -> String
prettyL (Id s) = prettyL0 "id" s
prettyL (Fix s p1 p2) = prettyL2 "Fix" s p1 p2
prettyL (PrePost s p1 p2) = prettyL2 "PrePost" s p1 p2
prettyL (Zero s p) = prettyL1 "Zero" s p
prettyL (ImpR s p) = prettyL1 "\\imp R" s p
prettyL (ImpL s p1 p2) = prettyL2 "\\imp L" s p1 p2
prettyL (NotL s p) = prettyL1 "\\lnot L" s p
prettyL (NotR s p) = prettyL1 "\\lnot R" s p
prettyL (WeakR s p) = prettyL1 "WeakR" s p
prettyL (AndR s p1 p2) = prettyL2 "\\land R" s p1 p2
prettyL (AndL s p) = prettyL1 "\\land L" s p
prettyL (OrL s p1 p2) = prettyL2 "\\lor L" s p1 p2
prettyL (OrR1 s p) = prettyL1 "\\lor R1" s p
prettyL (OrR2 s p) = prettyL1 "\\lor R2" s p
prettyL (BotL s) = prettyL0 "\\bot L" s
prettyL (TopR s) = prettyL0 "\\top R" s

prettyL0 :: String -> Sequent -> String
prettyL0 n s = "\\dfrac{}{" ++ prettyLS s ++ "} " ++ n

prettyL1 :: String -> Sequent -> Proof -> String
prettyL1 n s p = "\\dfrac{" ++ prettyL p ++ "}{" ++
                 prettyLS s ++ "} " ++ n

prettyL2 :: String -> Sequent -> Proof -> Proof -> String
prettyL2 n s p1 p2 = "\\dfrac{" ++ prettyL p1 ++ "\\hspace{5mm}" ++
                   prettyL p2 ++ "}{" ++ prettyLS s ++ "} " ++ n

prettyLR :: Result Proof -> String
prettyLR NoResult   = "NoResult"
prettyLR Timeout    = "Timeout"
prettyLR (Result x) = "Result (LaTeX follows):\n" ++ prettyL x
